@@ -7,12 +7,16 @@ import logging
 from abc import ABC, abstractmethod
 from datetime import datetime
 from functools import cached_property
-from typing import Any, Dict, Iterable, Mapping, MutableMapping, Optional, Union
+from typing import Any, Dict, Iterable, List, Mapping, MutableMapping, Optional, Union
 from urllib.parse import parse_qsl, urlparse
 
 import pendulum as pdm
 import requests
+from airbyte_cdk.sources.streams.core import StreamData
 from airbyte_cdk.sources.streams.http import HttpStream
+from airbyte_cdk.sources.streams.http.error_handlers import ErrorHandler, HttpStatusErrorHandler
+from airbyte_cdk.sources.streams.http.error_handlers.default_error_mapping import DEFAULT_ERROR_MAPPING
+from airbyte_protocol.models import SyncMode
 from requests.exceptions import RequestException
 from source_shopify.shopify_graphql.bulk.job import ShopifyBulkManager
 from source_shopify.shopify_graphql.bulk.query import ShopifyBulkQuery, ShopifyBulkTemplates
@@ -28,7 +32,7 @@ class ShopifyStream(HttpStream, ABC):
     logger = logging.getLogger("airbyte")
 
     # Latest Stable Release
-    api_version = "2023-07"
+    api_version = "2024-04"
     # Page size
     limit = 250
 
@@ -36,9 +40,6 @@ class ShopifyStream(HttpStream, ABC):
     order_field = "updated_at"
     filter_field = "updated_at_min"
 
-    raise_on_http_errors = True
-    max_retries = 5
-
     def __init__(self, config: Dict) -> None:
         super().__init__(authenticator=config["authenticator"])
         self._transformer = DataTypeEnforcer(self.get_json_schema())
@@ -108,15 +109,10 @@ class ShopifyStream(HttpStream, ABC):
                 record["shop_url"] = self.config["shop"]
                 yield self._transformer.transform(record)
 
-    def should_retry(self, response: requests.Response) -> bool:
+    def get_error_handler(self) -> Optional[ErrorHandler]:
         known_errors = ShopifyNonRetryableErrors(self.name)
-        status = response.status_code
-        if status in known_errors.keys():
-            setattr(self, "raise_on_http_errors", False)
-            self.logger.warning(known_errors.get(status))
-            return False
-        else:
-            return super().should_retry(response)
+        error_mapping = DEFAULT_ERROR_MAPPING | known_errors
+        return HttpStatusErrorHandler(self.logger, max_retries=5, error_mapping=error_mapping)
 
 
 class ShopifyDeletedEventsStream(ShopifyStream):
@@ -631,7 +627,6 @@ class IncrementalShopifyGraphQlBulkStream(IncrementalShopifyStream):
     filter_field = "updated_at"
     cursor_field = "updated_at"
     data_field = "graphql"
-    http_method = "POST"
 
     parent_stream_class: Optional[Union[ShopifyStream, IncrementalShopifyStream]] = None
 
@@ -641,14 +636,15 @@ class IncrementalShopifyGraphQlBulkStream(IncrementalShopifyStream):
         self.query = self.bulk_query(shop_id=config.get("shop_id"))
         # define BULK Manager instance
         self.job_manager: ShopifyBulkManager = ShopifyBulkManager(
-            session=self._session,
+            session=self._http_client._session,
             base_url=f"{self.url_base}{self.path()}",
             stream_name=self.name,
+            query=self.query,
+            job_termination_threshold=float(config.get("job_termination_threshold", 3600)),
+            # overide the default job slice size, if provided (it's auto-adjusted, later on)
+            job_size=config.get("bulk_window_in_days", 0.0),
         )
-        # overide the default job slice size, if provided (it's auto-adjusted, later on)
-        self.bulk_window_in_days = config.get("bulk_window_in_days")
-        if self.bulk_window_in_days:
-            self.job_manager.job_size = self.bulk_window_in_days
+
         # define Record Producer instance
         self.record_producer: ShopifyBulkRecord = ShopifyBulkRecord(self.query)
 
@@ -687,27 +683,6 @@ class IncrementalShopifyGraphQlBulkStream(IncrementalShopifyStream):
         """NOT USED FOR BULK OPERATIONS TO SAVE THE RATE LIMITS AND TIME FOR THE SYNC."""
         return None
 
-    def request_params(self, **kwargs) -> MutableMapping[str, Any]:
-        """
-        NOT USED FOR SHOPIFY BULK OPERARTIONS.
-        https://shopify.dev/docs/api/usage/bulk-operations/queries#write-a-bulk-operation
-        """
-        return {}
-
-    def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
-        """
-        NOT USED FOR SHOPIFY BULK OPERATIONS.
-        https://shopify.dev/docs/api/usage/bulk-operations/queries#write-a-bulk-operation
-        """
-        return None
-
-    def request_body_json(self, stream_slice: Optional[Mapping[str, Any]] = None, **kwargs) -> Mapping[str, Any]:
-        """
-        Override for _send_request CDK method to send HTTP request to Shopify BULK Operatoions.
-        https://shopify.dev/docs/api/usage/bulk-operations/queries#bulk-query-overview
-        """
-        return {"query": ShopifyBulkTemplates.prepare(stream_slice.get("query"))}
-
     def get_updated_state(
         self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]
     ) -> MutableMapping[str, Any]:
@@ -748,7 +723,7 @@ class IncrementalShopifyGraphQlBulkStream(IncrementalShopifyStream):
             return self.config.get("start_date")
 
     def emit_slice_message(self, slice_start: datetime, slice_end: datetime) -> None:
-        slice_size_message = f"Slice size: `P{round(self.job_manager.job_size, 1)}D`"
+        slice_size_message = f"Slice size: `P{round(self.job_manager._job_size, 1)}D`"
         self.logger.info(f"Stream: `{self.name}` requesting BULK Job for period: {slice_start} -- {slice_end}. {slice_size_message}")
 
     @stream_state_cache.cache_stream_state
@@ -761,21 +736,23 @@ class IncrementalShopifyGraphQlBulkStream(IncrementalShopifyStream):
                 self.job_manager.job_size_normalize(start, end)
                 slice_end = self.job_manager.get_adjusted_job_start(start)
                 self.emit_slice_message(start, slice_end)
-                yield {"query": self.query.get(self.filter_field, start.to_rfc3339_string(), slice_end.to_rfc3339_string())}
+                yield {"start": start.to_rfc3339_string(), "end": slice_end.to_rfc3339_string()}
                 # increment the end of the slice or reduce the next slice
                 start = self.job_manager.get_adjusted_job_end(start, slice_end)
         else:
             # for the streams that don't support filtering
-            yield {"query": self.query.get()}
+            yield {}
 
-    def process_bulk_results(
+    def read_records(
         self,
-        response: requests.Response,
+        sync_mode: SyncMode,
+        cursor_field: Optional[List[str]] = None,
+        stream_slice: Optional[Mapping[str, Any]] = None,
         stream_state: Optional[Mapping[str, Any]] = None,
-    ) -> Optional[Iterable[Mapping[str, Any]]]:
-        # process the CREATED Job prior to other actions
-        self.job_manager.job_process_created(response)
-        # get results fetched from COMPLETED BULK Job
+    ) -> Iterable[StreamData]:
+        self.job_manager.create_job(stream_slice, self.filter_field)
+        stream_state = stream_state_cache.cached_state.get(self.name, {self.cursor_field: self.default_state_comparison_value})
+
         filename = self.job_manager.job_check_for_completion()
         # the `filename` could be `None`, meaning there are no data available for the slice period.
         if filename:
@@ -785,8 +762,3 @@ class IncrementalShopifyGraphQlBulkStream(IncrementalShopifyStream):
                 self.record_producer.read_file(filename)
             )
             yield from self.filter_records_newer_than_state(stream_state, records)
-
-    def parse_response(self, response: requests.Response, **kwargs) -> Iterable[Mapping]:
-        # get the cached substream state, to avoid state collisions for Incremental Syncs
-        stream_state = stream_state_cache.cached_state.get(self.name, {self.cursor_field: self.default_state_comparison_value})
-        yield from self.process_bulk_results(response, stream_state)
